import enum
import functools
import re

import numpy as np
from karabo.bound import (
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    ChannelMetaData,
    Epochstamp,
    Hash,
    ImageData,
    MetricPrefix,
    Timestamp,
    Trainstamp,
    Unit,
)
from TrainMatcher import TrainMatcher

from . import geom_utils, preview_utils, scenes, schemas
from ._version import version as deviceVersion


corr_source_re = re.compile(r".*\/CORRECT(\d+)")
xtdf_source_re = re.compile(r".*\/DET\/(\d+)CH0:xtdf")
daq_source_re = re.compile(r".*\/DET\/.*?(\d+):daqOutput")


class BridgeOutputOptions(enum.Enum):
    MATCHED = "matched"
    ASSEMBLED = "assembled"


@KARABO_CLASSINFO("DetectorAssembler", deviceVersion)
class DetectorAssembler(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("availableScenes")
            .setNewDefaultValue(["overview", "trainMatcherScene"])
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("timeOfFlight")
            .displayedName("Time of flight")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .warnHigh(1000)
            .info("Time of flight exceeding 1 s")
            .needsAcknowledging(False)
            .commit(),

            STRING_ELEMENT(expected)
            .key("pathToStack")
            .assignmentOptional()
            .defaultValue("image.data")
            .commit(),

            OUTPUT_CHANNEL(expected)
            .key("assembledOutput")
            .dataSchema(schemas.preview_schema(wrap_image_in_imagedata=False))
            .commit(),

            STRING_ELEMENT(expected)
            .key("outputForBridgeOutput")
            .displayedName("Bridge output")
            .description(
                "The bridge can send the same data as any one of the three Karabo "
                "output channels. If this setting is 'matched', it will follow "
                "'output', if 'assembled', it will follow 'assembledOutput' and if "
                "'preview', it will follow 'preview.output'."
            )
            .options(",".join(option.value for option in BridgeOutputOptions))
            .assignmentOptional()
            .defaultValue("matched")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("geometryDevice")
            .displayedName("Geometry device")
            .description(
                "The the device which will provide geometry. The device is expected to "
                "have current geometry as a string (for now using custom non-robust "
                "serialization scheme) element named serializedGeometry."
            )
            .assignmentMandatory()
            .commit(),
        )
        preview_utils.PreviewFriend.add_schema(expected, "preview", create_node=True)
        (
            OVERWRITE_ELEMENT(expected)
            .key("preview.flipSS")
            .setNewDefaultValue(True)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("preview.flipFS")
            .setNewDefaultValue(True)
            .commit(),
        )

    def __init__(self, conf):
        super().__init__(conf)
        self.info.merge(Hash("timeOfFlight", 0))
        self.registerSlot(self.slotReceiveGeometry)

    def initialization(self):
        super().initialization()

        # TODO: match inside device, fill multiple independent buffers

        self._preview_friend = preview_utils.PreviewFriend(self)
        self._path_to_stack = self.get("pathToStack")
        self._geometry = None
        self._stack_input_buffer = None

        self.KARABO_SLOT(self.requestScene)

        self.signalSlotable.connect(
            self.get("geometryDevice"),
            "signalNewGeometry",
            "",  # slot device ID (default: self)
            "slotReceiveGeometry",
        )

        self.assembled_output = self.signalSlotable.getOutputChannel("assembledOutput")
        self.start()

    def requestScene(self, params):
        scene_name = params.get("name", default="")
        if scene_name == "trainMatcherScene":
            params["name"] = "scene"
            return super().requestScene(params)
        if scene_name != "overview":
            self.log.INFO(
                f"Unexpected scene name {scene_name} requested, "
                "will just return overview"
            )
        payload = Hash("name", scene_name, "success", True)
        payload["data"] = scenes.detector_assembler_overview(
            device_id=self.getInstanceId(),
            schema=self.getFullSchema(),
        )
        self.reply(
            Hash(
                "type",
                "deviceScene",
                "origin",
                self.getInstanceId(),
                "payload",
                payload,
            )
        )

    def slotReceiveGeometry(self, device_id, serialized_geometry):
        self.log.INFO(f"Received geometry from {device_id}")
        try:
            self._geometry = geom_utils.deserialize_geometry(serialized_geometry)
        except Exception as e:
            self.log.WARN(f"Failed to deserialize geometry; {e}")
        # TODO: allow multiple memory cells (extra geom notion of extra dimensions)
        self._stack_input_buffer = np.ma.masked_array(
            data=np.zeros(self._geometry.expected_data_shape, dtype=np.float32),
            mask=False,
        )
        self._assemble_buffer = np.ma.masked_array(
            data=self._geometry.output_array_for_position_fast(),
            mask=False,
        )

    def on_matched_data(self, train_id, sources):
        if self._geometry is None:
            self.log.WARN("Have not received a geometry yet, will not send anything")
            return
        my_timestamp = Timestamp(Epochstamp(), Trainstamp(train_id))
        my_device_id = self.getInstanceId()
        bridge_output_choice = BridgeOutputOptions(
            self.unsafe_get("outputForBridgeOutput")
        )

        module_indices_unfilled = set(range(self._stack_input_buffer.shape[0]))
        earliest_source_timestamp = float("inf")
        self._stack_input_buffer.mask.fill(False)
        for source, (data, source_timestamp) in sources.items():
            # regular TrainMatcher output
            self.output.write(
                data, ChannelMetaData(source, source_timestamp), copyAllData=False
            )
            if bridge_output_choice is BridgeOutputOptions.MATCHED:
                self.zmq_output.write(source, data, source_timestamp)

            # prepare for assembly
            # TODO: handle failure to "parse" source, get data out
            image_data = data["image.data"]
            if isinstance(image_data, ImageData):
                # TODO: maybe glance encoding here
                image_data = image_data.getData()
            image_data = image_data.astype(np.float32, copy=False)  # TODO: set dtype based on input?
            if data.has("image.mask"):
                image_mask = data["image.mask"]
                if isinstance(image_mask, ImageData):
                    image_mask = image_mask.getData()
            else:
                image_mask = False
            module_index = self._source_to_index(source)
            masked = np.ma.masked_array(
                data=image_data,
                mask=image_mask,
            )
            self._stack_input_buffer[module_index] = masked
            module_indices_unfilled.discard(module_index)
            earliest_source_timestamp = min(
                earliest_source_timestamp, source_timestamp.toTimestamp()
            )

        self.output.update()
        if bridge_output_choice is BridgeOutputOptions.MATCHED:
            self.zmq_output.update()

        for module_index in module_indices_unfilled:
            self._stack_input_buffer.mask[module_index].fill(True)
            # consider configurable treatment of missing modules

        assembled, _ = self._geometry.position_modules_fast(
            self._stack_input_buffer, out=self._assemble_buffer
        )
        output_hash = Hash(
            "image.data",
            assembled,
            "image.mask",
            assembled.mask,
        )
        self.assembled_output.write(
            output_hash,
            ChannelMetaData(f"{my_device_id}:assembledOutput", my_timestamp),
            copyAllData=False,
        )
        self.assembled_output.update()
        if bridge_output_choice is BridgeOutputOptions.ASSEMBLED:
            self.zmq_output.write(
                f"{my_device_id}:assembledOutput", output_hash, my_timestamp
            )
            self.zmq_output.update()

        self._preview_friend.write_outputs(
            my_timestamp,
            assembled,
        )

        self.info["timeOfFlight"] = (
            Timestamp().toTimestamp() - earliest_source_timestamp
        ) * 1000
        self.info["sent"] += 1
        self.info["trainId"] = train_id
        self.rate_out.update()

    def on_new_data(self, channel, data, meta):
        super().on_new_data(channel, data, meta)
        ...
        # TODO: start filling buffer early

    @functools.lru_cache()
    def _source_to_index(self, source):
        # note: cache means warning only shows up once (also not performance-critical)
        # TODO: allow user to inspect, modify the mapping

        if (match := corr_source_re.match(source)) is not None:
            return int(match.group(1))

        if (match := xtdf_source_re.match(source)) is not None:
            return int(match.group(1))

        if (match := daq_source_re.match(source)) is not None:
            return int(match.group(1)) - 1

        self.log.WARN(f"Couldn't figure out index for source {source}")
        return 0

    def preReconfigure(self, conf):
        super().preReconfigure(conf)
        self._preview_friend.reconfigure(conf)


# forward-compatible unsafe_get proposed by @haufs
if not hasattr(DetectorAssembler, "unsafe_get"):

    def unsafe_get(self, key):
        """See base_correction.py"""
        return self._parameters.get(key)

    setattr(DetectorAssembler, "unsafe_get", unsafe_get)
