from timeit import default_timer

import numpy as np
from karabo.bound import (
    DOUBLE_ELEMENT,
    INPUT_CHANNEL,
    KARABO_CLASSINFO,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    UINT64_ELEMENT,
    VECTOR_UINT32_ELEMENT,
    VECTOR_STRING_ELEMENT,
    ChannelMetaData,
    Hash,
    MetricPrefix,
    PythonDevice,
    State,
    Timestamp,
    Unit,
)
from karabo.common.api import KARABO_SCHEMA_DISPLAY_TYPE_SCENES as DT_SCENES

from . import scenes, schemas, shmem_utils, utils
from .base_correction import WarningLampType
from ._version import version as deviceVersion

# plan:
# - split data into 8 modules
# - copy each into shared memory and send on distinct channel

PROCESSING_STATE_TIMEOUT = 10


@KARABO_CLASSINFO("LpdminiSplitter", deviceVersion)
class LpdminiSplitter(PythonDevice):
    # up to 8 minis, using 1-indexing for naming
    _output_channel_names = [f"output-{i+1}" for i in range(8)]

    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("state")
            .setNewDefaultValue(State.INIT)
            .commit(),

            INPUT_CHANNEL(expected)
            .key("input")
            .commit(),

            VECTOR_UINT32_ELEMENT(expected)
            .key("outputDataShape")
            .displayedName("Output data shape")
            .readOnly()
            .initialValue([512, 32, 256])
            .commit(),

            UINT64_ELEMENT(expected)
            .key("trainId")
            .displayedName("Train ID")
            .description("ID of latest train processed by this device.")
            .readOnly()
            .initialValue(0)
            .commit(),

            NODE_ELEMENT(expected)
            .key("performance")
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("performance.rate")
            .displayedName("Rate")
            .description(
                "Actual rate with which this device gets, processes, and sends trains. "
                "This is a simple windowed moving average."
            )
            .unit(Unit.HERTZ)
            .readOnly()
            .initialValue(0)
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("performance.processingTime")
            .displayedName("Processing time")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .warnHigh(100)
            .info("Processing too slow to reach 10 Hz")
            .needsAcknowledging(False)
            .commit(),

            VECTOR_STRING_ELEMENT(expected)
            .key("availableScenes")
            .setSpecialDisplayType(DT_SCENES)
            .readOnly()
            .initialValue(["overview"])
            .commit(),
        )

        for channel_name in LpdminiSplitter._output_channel_names:
            (
                OUTPUT_CHANNEL(expected)
                .key(channel_name)
                .dataSchema(schemas.xtdf_output_schema())
                .commit(),
            )

    def __init__(self, config):
        super().__init__(config)
        self.registerInitialFunction(self._initialization)
        self._output_image_shape = (512, 32, 256)  # will be updated based on frames
        self.KARABO_SLOT(self.requestScene)
        self._latest_log_message = None
        self._latest_warn_type = None

    def _initialization(self):
        self.KARABO_ON_DATA("input", self.input_handler)
        self._shmem_buffers = [
            shmem_utils.ShmemCircularBuffer(
                1 * 2**30,
                (512, 32, 256),
                np.uint16,
                f"{self.getInstanceId()}:{channel_name}"
            )
            for channel_name in self._output_channel_names
        ]

        # performance measures and such
        self._last_processing_started = 0  # used for processing time and timeout
        self._buffered_status_update = Hash()
        self._processing_time_tracker = utils.ExponentialMovingAverage(alpha=0.3)
        self._rate_tracker = utils.WindowRateTracker()
        self._input_delay_tracker = utils.ExponentialMovingAverage(alpha=0.3)
        self._performance_measure_update_timer = utils.RepeatingTimer(
            interval=1,
            callback=self._update_performance_measures,
        )
        self.updateState(State.ON)

    def input_handler(self, data_hash, metadata):
        state = self.get("state")
        if state is State.INIT:
            return

        # TODO: handle empty DAQ hash
        if not data_hash.has("image.data"):
            self.log_status_info("Input hash had no image data node")
            if state is State.PROCESSING:
                self.updateState(State.IGNORING)
            return
        try:
            image_data = np.asarray(
                data_hash.get("image.data")
            )
        except RuntimeError as err:
            self.log_status_info(
                f"Failed to load image data; probably empty hash from DAQ: {err}",
                WarningLampType.EMPTY_HASH,
            )
            if state is State.PROCESSING:
                self.updateState(State.IGNORING)
            return

        if state is not State.PROCESSING:
            self.log_status_info("Processing data")
            self.updateState(State.PROCESSING)

        self._last_processing_started = default_timer()
        timestamp = Timestamp.fromHashAttributes(metadata.getAttributes("timestamp"))
        self._buffered_status_update.set("trainId", timestamp.getTrainId())
        num_frames = data_hash["image.cellId"].size
        image_data = image_data.reshape((num_frames, 256, 256))
        # note: reshape is equivalent to overrideInputAxisOrder in BaseCorrection
        if num_frames != self._output_image_shape:
            self._output_image_shape = (num_frames, 32, 256)
            for shmem_buffer in self._shmem_buffers:
                shmem_buffer.change_shape(self._output_image_shape)
            self.set("outputDataShape", list(self._output_image_shape))
        # reshape to drop singleton axis + optionally fix DAQ shape shenanigans
        receiver_source_name = metadata.get("source")
        for i, (channel_name, shmem_buffer) in enumerate(
            zip(self._output_channel_names, self._shmem_buffers)
        ):
            this_slice = image_data[:, i * 32 : (i + 1) * 32]
            buffer_handle, buffer_array = shmem_buffer.next_slot()
            buffer_array[:] = this_slice
            data_hash.set("image.data", buffer_handle)
            data_hash.set("calngShmemPaths", ["image.data"])
            # TODO: consider source name; virtual indices like in offline?
            channel = self.signalSlotable.getOutputChannel(channel_name)
            channel.write(
                data_hash,
                ChannelMetaData(f"{receiver_source_name}-{i+1}", timestamp),
                # adding 1-indexed suffix index to soruce name forwarded
                copyAllData=False,
            )
            channel.update()
        self._processing_time_tracker.update(
            default_timer() - self._last_processing_started
        )
        self._rate_tracker.update()

    def _update_performance_measures(self):
        if self.get("state") in {State.PROCESSING, State.IGNORING}:
            self._buffered_status_update.set(
                "performance.rate", self._rate_tracker.get()
            )
            self._buffered_status_update.set(
                "performance.processingTime", self._processing_time_tracker.get() * 1000
            )
            self.set(self._buffered_status_update)
            if (
                default_timer() - self._last_processing_started
                > PROCESSING_STATE_TIMEOUT
            ):
                self.updateState(State.ON)

    def requestScene(self, params):
        payload = Hash()
        payload["name"] = "overview"
        payload["data"] = scenes.lpdmini_splitter_overview(
            device_id=self.getInstanceId(),
            schema=self.getFullSchema(),
        )
        payload["success"] = True
        response = Hash()
        response["type"] = "deviceScene"
        response["origin"] = self.getInstanceId()
        response["payload"] = payload
        self.reply(response)

    def log_status_info(self, msg, warn_type=None):
        if (msg != self._latest_log_message) and (
            warn_type is None or warn_type != self._latest_warn_type
        ):
            self.log.INFO(msg)
            self.set("status", msg)
            self._latest_log_message = msg
            self._latest_warn_type = warn_type
