import concurrent.futures
import enum
import re

import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    INT32_ELEMENT,
    KARABO_CLASSINFO,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    TABLE_ELEMENT,
    ChannelMetaData,
    Hash,
    Schema,
    State,
    VectorHash,
)
from TrainMatcher import TrainMatcher

from . import shmem_utils, utils
from ._version import version as deviceVersion


class GroupType(enum.Enum):
    MULTISOURCE = "sources"  # same key stacked from multiple sources in new source
    MULTIKEY = "keys"  # multiple keys within each matched source is stacked in new key


class MergeMethod(enum.Enum):
    STACK = "stack"
    INTERLEAVE = "interleave"


def merge_schema():
    schema = Schema()
    (
        BOOL_ELEMENT(schema)
        .key("select")
        .displayedName("Select")
        .assignmentOptional()
        .defaultValue(False)
        .reconfigurable()
        .commit(),

        STRING_ELEMENT(schema)
        .key("sourcePattern")
        .displayedName("Source pattern")
        .assignmentOptional()
        .defaultValue("")
        .reconfigurable()
        .commit(),

        STRING_ELEMENT(schema)
        .key("keyPattern")
        .displayedName("Key pattern")
        .assignmentOptional()
        .defaultValue("")
        .reconfigurable()
        .commit(),

        STRING_ELEMENT(schema)
        .key("replacement")
        .displayedName("Replacement")
        .assignmentOptional()
        .defaultValue("")
        .reconfigurable()
        .commit(),

        STRING_ELEMENT(schema)
        .key("groupType")
        .displayedName("Group type")
        .options(",".join(option.value for option in GroupType))
        .assignmentOptional()
        .defaultValue(GroupType.MULTISOURCE.value)
        .reconfigurable()
        .commit(),

        STRING_ELEMENT(schema)
        .key("mergeMethod")
        .displayedName("Merge method")
        .options(",".join(option.value for option in MergeMethod))
        .assignmentOptional()
        .defaultValue(MergeMethod.STACK.value)
        .reconfigurable()
        .commit(),

        INT32_ELEMENT(schema)
        .key("axis")
        .displayedName("Axis")
        .assignmentOptional()
        .defaultValue(0)
        .reconfigurable()
        .commit(),
    )

    return schema


@KARABO_CLASSINFO("ShmemTrainMatcher", deviceVersion)
class ShmemTrainMatcher(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            TABLE_ELEMENT(expected)
            .key("merge")
            .displayedName("Array stacking")
            .allowedStates(State.PASSIVE)
            .description(
                "Specify which source(s) or key(s) to stack or interleave."
                "When stacking sources, the 'Source pattern' is interpreted as a "
                "regular expression and the 'Key pattern' is interpreted as an "
                "ordinary string. From all sources matching the source pattern, the "
                "data under this key (should be array with same dimensions across all "
                "stacked sources) is stacked in the same order as the sources are "
                "listed in 'Data sources' and the result is under the same key name in "
                "a new source named by 'Replacement'. "
                "When stacking keys, both the 'Source pattern' and the 'Key pattern' "
                "are regular expressions. Within each source matching the source "
                "pattern, all keys matching the key pattern are stacked and the result "
                "is put under the key named by 'Replacement'. "
                "While source stacking is optimized and can use thread pool, key "
                "stacking will iterate over all paths in matched sources and naively "
                "call np.stack for each key pattern. In either case, data that is used "
                "for stacking is removed from its original location (e.g. key is "
                "erased from hash). "
                "In both cases, the data can alternatively be interleaved. This is "
                "essentially equivalent to stacking except followed by a reshape such "
                "that the output is shaped like concatenation."
            )
            .setColumns(merge_schema())
            .assignmentOptional()
            .defaultValue([])
            .reconfigurable()
            .commit(),

            # order is important for stacking, disable sorting
            OVERWRITE_ELEMENT(expected)
            .key("sortSources")
            .setNowReadOnly()
            .setNewDefaultValue(False)
            .commit(),

            BOOL_ELEMENT(expected)
            .key("useThreadPool")
            .displayedName("Use thread pool")
            .allowedStates(State.PASSIVE)
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected)
            .key("enableKaraboOutput")
            .displayedName("Enable Karabo channel")
            .allowedStates(State.PASSIVE)
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected)
            .key("useInfiniband")
            .description(
                "If enabled, device will during initialization try to bind its data "
                "output channel (output) to its node's infiniband interface. Default "
                "interface is used if no 'ib0' interface is found."
            )
            .assignmentOptional()
            .defaultValue(True)
            .commit(),
        )

    def __init__(self, config):
        if config.get("useInfiniband", default=True):
            from PipeToZeroMQ.utils import find_infiniband_ip
            config["output.hostname"] = find_infiniband_ip()
        super().__init__(config)

    def initialization(self):
        self._stacking_buffers = {}
        self._source_stacking_indices = {}
        self._source_stacking_sources = {}
        self._source_stacking_group_sizes = {}
        self._key_stacking_sources = {}
        self._have_prepared_merge_groups = False
        self._prepare_merge_groups()
        super().initialization()
        self._shmem_handler = shmem_utils.ShmemCircularBufferReceiver()
        if self.get("useThreadPool"):
            self._thread_pool = concurrent.futures.ThreadPoolExecutor()
        else:
            self._thread_pool = None

        if not self.get("enableKaraboOutput"):
            # it is set already by super by default, so only need to turn off
            self.output = None

    def preReconfigure(self, conf):
        super().preReconfigure(conf)
        if conf.has("merge") or conf.has("sources"):
            self._have_prepared_merge_groups = False
            # re-prepare in postReconfigure after sources *and* merge are in self
        if conf.has("useThreadPool"):
            if self._thread_pool is not None:
                self._thread_pool.shutdown()
                self._thread_pool = None
            if conf["useThreadPool"]:
                self._thread_pool = concurrent.futures.ThreadPoolExecutor()
        if conf.has("enableKaraboOutput"):
            if conf["enableKaraboOutput"]:
                self.output = self._ss.getOutputChannel("output")
            else:
                self.output = None

    def postReconfigure(self):
        super().postReconfigure()
        if not self._have_prepared_merge_groups:
            self._prepare_merge_groups()

    def _prepare_merge_groups(self):
        # not filtering by row["select"] to allow unselected sources to create gaps
        source_names = [row["source"].partition("@")[0] for row in self.get("sources")]
        self._stacking_buffers.clear()
        self._source_stacking_indices.clear()
        self._source_stacking_sources.clear()
        self._source_stacking_group_sizes.clear()
        self._key_stacking_sources.clear()
        # split by type, prepare regexes
        for row in self.get("merge"):
            if not row["select"]:
                continue
            group_type = GroupType(row["groupType"])
            source_re = re.compile(row["sourcePattern"])
            merge_method = MergeMethod(row["mergeMethod"])
            axis = row["axis"]
            if group_type is GroupType.MULTISOURCE:
                key = row["keyPattern"]
                new_source = row["replacement"]
                merge_sources = [
                    source for source in source_names if source_re.match(source)
                ]
                if len(merge_sources) == 0:
                    self.log.WARN(
                        f"Group pattern {source_re} did not match any known sources"
                    )
                    continue
                for (i, source) in enumerate(merge_sources):
                    self._source_stacking_sources.setdefault(source, []).append(
                        (key, new_source, merge_method, axis)
                    )
                    self._source_stacking_indices[(source, new_source, key)] = i
                self._source_stacking_group_sizes[(new_source, key)] = i + 1
            else:
                key_re = re.compile(row["keyPattern"])
                new_key = row["replacement"]
                self._key_stacking_sources.setdefault(source, []).append(
                    (key_re, new_key, merge_method, axis)
                )

        self._have_prepared_merge_groups = True

    def _update_stacking_buffer(
        self, new_source, key, individual_shape, merge_method, axis, dtype
    ):
        if merge_method is MergeMethod.STACK:
            self._stacking_buffers[(new_source, key)] = np.empty(
                shape=utils.stacking_buffer_shape(
                    individual_shape,
                    self._source_stacking_group_sizes[(new_source, key)],
                    axis=axis,
                ),
                dtype=dtype,
            )
        else:
            self._stacking_buffers[(new_source, key)] = np.empty(
                shape=utils.interleaving_buffer_shape(
                    individual_shape,
                    self._source_stacking_group_sizes[(new_source, key)],
                    axis=axis,
                ),
                dtype=dtype,
            )

    def _handle_source(self, source, data_hash, timestamp, new_sources_map):
        # dereference calng shmem handles
        self._shmem_handler.dereference_shmem_handles(data_hash)

        # stack across sources (many sources, same key)
        # could probably save ~100 ns by "if ... in" instead of get
        for (
            key,
            new_source,
            merge_method,
            axis,
        ) in self._source_stacking_sources.get(source, ()):
            merge_data = data_hash.get(key)
            merge_index = self._source_stacking_indices[
                (source, new_source, key)
            ]
            try:
                merge_buffer = self._stacking_buffers[(new_source, key)]
                utils.set_on_axis(
                    merge_buffer,
                    merge_data,
                    merge_index
                    if merge_method is MergeMethod.STACK
                    else np.index_exp[
                        slice(
                            merge_index,
                            None,
                            self._source_stacking_group_sizes[(new_source, key)],
                        )
                    ],
                    axis,
                )
            except (ValueError, IndexError, KeyError):
                # ValueError: wrong shape (react to merge_data.shape)
                # KeyError: buffer doesn't exist yet
                # IndexError: new source? (TODO: re-run _prepare_merge_groups or ERROR)
                # either way, create appropriate buffer now
                # TODO: complain if shape varies between sources within train
                self._update_stacking_buffer(
                    new_source,
                    key,
                    merge_data.shape,
                    merge_method,
                    axis=axis,
                    dtype=merge_data.dtype,
                )
                # and then try again
                merge_buffer = self._stacking_buffers[(new_source, key)]
                utils.set_on_axis(
                    merge_buffer,
                    merge_data,
                    merge_index
                    if merge_method is MergeMethod.STACK
                    else np.index_exp[
                        slice(
                            merge_index,
                            None,
                            self._source_stacking_group_sizes[(new_source, key)],
                        )
                    ],
                    axis,
                )
            # TODO: zero out unfilled buffer entries
            data_hash.erase(key)

            if new_source not in new_sources_map:
                new_sources_map[new_source] = (Hash(), timestamp)
            new_source_hash = new_sources_map[new_source][0]
            if not new_source_hash.has(key):
                new_source_hash[key] = merge_buffer

        # stack keys (multiple keys within this source)
        for (key_re, new_key, merge_method, axis) in self._key_stacking_sources.get(
            source, ()
        ):
            # note: please no overlap between different key_re
            # note: if later key_re match earlier new_key, this gets spicy
            keys = [key for key in data_hash.paths() if key_re.match(key)]
            try:
                # note: maybe we could reuse buffers here, too?
                if merge_method is MergeMethod.STACK:
                    stacked = np.stack([data_hash.get(key) for key in keys], axis=axis)
                else:
                    first = data_hash.get(keys[0])
                    stacked = np.empty(
                        shape=utils.stacking_buffer_shape(
                            first.shape, len(keys), axis=axis
                        ),
                        dtype=first.dtype,
                    )
                    for i, key in enumerate(keys):
                        utils.set_on_axis(
                            stacked,
                            data_hash.get(key),
                            np.index_exp[slice(i, None, len(keys))],
                            axis,
                        )
            except Exception as e:
                self.log.WARN(f"Failed to stack {key_re} for {source}: {e}")
            else:
                for key in keys:
                    data_hash.erase(key)
                data_hash[new_key] = stacked

    def on_matched_data(self, train_id, sources):
        new_sources_map = {}
        if self._thread_pool is None:
            for source, (data, timestamp) in sources.items():
                self._handle_source(source, data, timestamp, new_sources_map)
        else:
            concurrent.futures.wait(
                [
                    self._thread_pool.submit(
                        self._handle_source, source, data, timestamp, new_sources_map
                    )
                    for source, (data, timestamp) in sources.items()
                ]
            )
        sources.update(new_sources_map)

        # karabo output
        if self.output is not None:
            for source, (data, timestamp) in sources.items():
                self.output.write(
                    data, ChannelMetaData(source, timestamp), copyAllData=False
                )
            self.output.update()

        # karabo bridge output
        for source, (data, timestamp) in sources.items():
            self.zmq_output.write(source, data, timestamp)
        self.zmq_output.update()

        self.info["sent"] += 1
        self.info["trainId"] = train_id
        self.rate_out.update()
