import enum

from karabo.middlelayer import (
    Assignment,
    AccessMode,
    AccessLevel,
    Bool,
    DaqPolicy,
    Configurable,
    Device,
    Hash,
    Proxy,
    Slot,
    State,
    String,
    VectorHash,
    VectorString,
    background,
    connectDevice,
    disconnectDevice,
    getConfiguration,
    get_property,
    slot,
    waitUntilNew,
)
from ._version import version as deviceVersion
from . import scenes


class PipelineOperationMode(enum.Enum):
    MANAGED = enum.auto()
    STANDALONE = enum.auto()


class KeyMappingRow(Configurable):
    managerKey = String(defaultValue="")
    controlValue = String(defaultValue="")
    idealValue = String(defaultValue="")
    managerValue = String(defaultValue="")
    matches = String(defaultValue="NORMAL", displayType="State")


class ConditionBase(Device):
    __version__ = deviceVersion
    managerDeviceId = String(
        displayedName="Manager device ID",
        assignment=Assignment.MANDATORY,
        accessMode=AccessMode.INITONLY,
    )

    keyMapping = VectorHash(
        displayedName="Key mapping",
        description="Read-only table used to show which properties go where.",
        rows=KeyMappingRow,
        accessMode=AccessMode.READONLY,
    )

    conditionsMatch = String(displayType="State")

    @property
    def keys_to_get(self):
        """Must return dict mapping device IDs to tuples of (device ID key, correction
        constant condition parameter key, [optional, can be None] function to
        translate control device value into condition parameter value.
        """
        raise NotImplementedError("Subclass must implement!")

    async def onInitialization(self):
        self.state = State.INIT
        self._manager_dev = await connectDevice(self.managerDeviceId.value)
        if self._manager_dev.classId.value == "CalibrationManager":
            self._operation_mode = PipelineOperationMode.MANAGED
        else:
            self._operation_mode = PipelineOperationMode.STANDALONE
        self.state = State.ON

    def _trigger_constant_loading(self):
        if self._operation_mode is PipelineOperationMode.MANAGED:
            background(self._manager_dev.managedKeys.loadMostRecentConstants())
        else:
            background(self._manager_dev.loadMostRecentConstants())

    @property
    def _manager_parameters_node(self):
        if self._operation_mode is PipelineOperationMode.MANAGED:
            return self._manager_dev.managedKeys.constantParameters
        else:
            return self._manager_dev.constantParameters

    updateManagerOnMonitor = Bool(
        displayedName="Monitor: update manager",
        description="Whenever parameter changes on monitored control devices, "
        "automatically reflect changes in operating conditions set on manager.",
        defaultValue=True,
        assignment=Assignment.OPTIONAL,
        accessMode=AccessMode.RECONFIGURABLE,
    )

    loadConstantsOnMonitor = Bool(
        displayedName="Monitor: load constants",
        description="After automatically reflecting control device parameter settings "
        "on manager, automatically trigger loading of most recent constants matching "
        "these parameters. Should probably only be used in conjunction with "
        "updateManagerOnMonitor.",
        defaultValue=False,
        assignment=Assignment.OPTIONAL,
        accessMode=AccessMode.RECONFIGURABLE,
    )

    @Slot(
        allowedStates=[State.ON],
        displayedName="Start monitoring",
        description="Will keep settings on manager in sync with control device(s) "
        "automagically by monitoring for updates.",
    )
    async def startMonitoring(self):
        self.state = State.CHANGING
        control_devs = {
            control_id: await connectDevice(control_id)
            for control_id in self.keys_to_get.keys()
        }

        # first bring in line
        if (
            await self._check_or_update(
                update_manager=self.updateManagerOnMonitor.value,
                control_devs=control_devs,
            )
            and self.loadConstantsOnMonitor.value
        ):
            self._trigger_constant_loading()

        # then start monitoring
        async def aux():
            while True:
                await waitUntilNew(
                    *(
                        get_property(control_devs[control_id], control_key)
                        for control_id, v in self.keys_to_get.items()
                        for control_key, *_ in v  # "v" for "variable naming is hard"
                    )
                )
                await self._check_or_update(
                    self.updateManagerOnMonitor.value, control_devs=control_devs
                )
                # TODO: maybe debounce
                if self.loadConstantsOnMonitor.value:
                    self._trigger_constant_loading()

        self._monitor = background(aux)
        self._control_devs = control_devs
        self.state = State.MONITORING

    @Slot(
        allowedStates=[State.MONITORING],
        displayedName="Stop monitoring",
    )
    async def stopMonitoring(self):
        self.state = State.CHANGING
        self._monitor.cancel()
        for dev in self._control_devs.values():
            await disconnectDevice(dev)
        self.state = State.ON

    async def _check_or_update(self, update_manager, control_devs=None):
        """Will compare values of keys_to_get on control device(s) to actual values on
        manager. Optionally changing manager settings accordingly. Optionally provide
        through control_devs a dict mapping control device ID to control device
        configurations or proxies.
        """
        manager_params_checked = set()
        key_mapping = []
        happy = True
        updated_manager = False
        for control_id, to_get in self.keys_to_get.items():
            if control_devs is None:
                control_dev = await getConfiguration(control_id)
            else:
                control_dev = control_devs[control_id]

            for control_key, manager_key, translator in to_get:
                (
                    control_value,
                    ideal_value,
                    manager_value,
                ), could_look_up = self._look_up(
                    control_dev, control_key, manager_key, translator
                )
                row_state = State.ON
                if not could_look_up:
                    happy = False
                    row_state = State.UNKNOWN
                elif ideal_value != manager_value:
                    if update_manager:
                        try:
                            setattr(
                                self._manager_parameters_node,
                                manager_key,
                                ideal_value,
                            )
                        except Exception as ex:
                            manager_value = f"Failed to set {ideal_value}: {ex}"
                            row_state = State.ERROR
                        else:
                            manager_value = ideal_value
                            updated_manager = True
                    else:
                        happy = False
                        row_state = State.ERROR
                key_mapping.append(
                    (
                        manager_key,
                        control_value,
                        ideal_value,
                        manager_value,
                        row_state.value,
                    )
                )
                manager_params_checked.add(manager_key)

        for unchecked_parameter in (
            set(dir(self._manager_parameters_node)) - manager_params_checked
        ):
            key_mapping.append(
                (
                    unchecked_parameter,
                    "",
                    "",
                    get_property(
                        self._manager_parameters_node, unchecked_parameter
                    ).value,
                    State.IGNORING.value,
                )
            )

        self.keyMapping = key_mapping
        if happy:
            self.conditionsMatch = State.ON.value
        else:
            self.conditionsMatch = State.ERROR.value

        return updated_manager

    def _look_up(
        self,
        control_dev,
        control_key,
        manager_key,
        translator,
    ):
        could_look_up = True
        if isinstance(control_dev, Proxy):
            # device proxy via connectDevice
            try:
                control_value = get_property(control_dev, control_key).value
            except AttributeError:
                control_value = "key not found"
                could_look_up = False
        else:
            # device configuration hash via getConfiguration
            if control_dev.has(control_key):
                control_value = control_dev[control_key]
            else:
                control_value = "key not found"
                could_look_up = False

        if could_look_up:
            try:
                ideal_value = (
                    control_value if translator is None else translator(control_value)
                )
            except Exception as ex:
                warning = f"Failed to process control value {control_value}; {ex}"
                ideal_value = warning
                self.log.WARN(warning)
                could_look_up = False
        else:
            ideal_value = "control key not found"

        try:
            manager_value = get_property(
                self._manager_parameters_node, manager_key
            ).value
        except AttributeError:
            manager_value = "key not found"
            could_look_up = False

        return (control_value, ideal_value, manager_value), could_look_up

    @Slot(
        allowedStates=[State.ON],
        displayedName="Check conditions",
        description="Compares current settings on manager to control device(s), but "
        "does not change anything on manager.",
    )
    async def checkConditions(self):
        self.state = State.CHANGING
        await self._check_or_update(False)
        self.state = State.ON

    @Slot(
        allowedStates=[State.ON],
        displayedName="Copy conditions",
        description="Copies current settings from control device(s) to manager. See "
        "also 'Start monitoring' for automatic version of this.",
    )
    async def updateConditions(self):
        self.state = State.CHANGING
        await self._check_or_update(True)
        self.state = State.ON

    availableScenes = VectorString(
        displayedName="Available scenes",
        displayType="Scenes",
        requiredAccessLevel=AccessLevel.OBSERVER,
        accessMode=AccessMode.READONLY,
        defaultValue=[
            "overview",
        ],
        daqPolicy=DaqPolicy.OMIT,
    )

    @slot
    def requestScene(self, params):
        name = params.get("name", default="overview")
        if name == "overview":
            scene_data = scenes.condition_checker_overview(
                self.deviceId,
                self.getDeviceSchema(),
            )
            payload = Hash("success", True, "name", name, "data", scene_data)

        return Hash("type", "deviceScene", "origin", self.deviceId, "payload", payload)
