import contextlib
import logging

import matplotlib.pyplot as plt
import numpy as np
from karabo.middlelayer import (
    AccessLevel,
    AccessMode,
    Assignment,
    Bool,
    Configurable,
    DaqPolicy,
    Device,
    Double,
    EncodingType,
    Hash,
    Image,
    ImageData,
    Int32,
    Node,
    Signal,
    Slot,
    State,
    String,
    Unit,
    VectorHash,
    VectorString,
    sleep,
    slot,
)
from matplotlib.backends.backend_agg import FigureCanvasAgg

from ._version import version as deviceVersion
from . import geom_utils, scenes


def make_x_y_coordinate_node(
    default_x, default_y, x_args=None, y_args=None, node_args=None
):
    class XYCoordinate(Configurable):
        x = Double(
            defaultValue=default_x,
            accessMode=AccessMode.RECONFIGURABLE,
            assignment=Assignment.OPTIONAL,
            **({} if x_args is None else x_args),
        )
        y = Double(
            defaultValue=default_y,
            accessMode=AccessMode.RECONFIGURABLE,
            assignment=Assignment.OPTIONAL,
            **({} if y_args is None else y_args),
        )

    return Node(XYCoordinate, **({} if node_args is None else node_args))


def make_x_y_offset_node():
    return make_x_y_coordinate_node(
        0,
        0,
        x_args={"unitSymbol": Unit.METER},
        y_args={"unitSymbol": Unit.METER},
        node_args={
            "displayedName": "Offset",
            "description": "See EXtra-geom documentation for details. This offset is "
            "applied to entire detector after initial geometry is created from manual "
            "parameters. Example: To move entire geometry up by 2 mm relative to beam, "
            "set offset.y to 2e-3.",
        },
    )


# TODO: consider other history models (could be fun)
class TweakGeometryNode(Configurable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._undo_stack = []
        self._redo_stack = []
        self._device = self.get_root()

    def _reset(self):
        # clear history when new geometry is set from manual / file
        self._undo_stack.clear()
        self._redo_stack.clear()
        self.undoLength = 0
        self.redoLength = 0

    @Slot(displayedName="Undo")
    async def undo(self):
        assert len(self._undo_stack) > 0
        self._redo_stack.append(self._device.geometry)
        await self._device._set_geometry(self._undo_stack.pop())
        self.undoLength = len(self._undo_stack)
        self.redoLength = len(self._redo_stack)

    @Slot(displayedName="Redo")
    async def redo(self):
        assert len(self._redo_stack) > 0
        self._undo_stack.append(self._device.geometry)
        await self._device._set_geometry(self._redo_stack.pop())
        self.undoLength = len(self._undo_stack)
        self.redoLength = len(self._redo_stack)

    offset = make_x_y_offset_node()

    @Slot(displayedName="Add offset")
    async def add(self):
        current_geometry = self._device.geometry
        new_geometry = current_geometry.offset(
            (self.offset.x.value, self.offset.y.value)
        )
        self._undo_stack.append(current_geometry)
        self.undoLength = len(self._undo_stack)
        await self._device._set_geometry(new_geometry)
        if self._redo_stack:
            self._redo_stack.clear()
            self.redoLength = 0

    undoLength = Int32(
        displayedName="Undo length",
        defaultValue=0,
        accessMode=AccessMode.READONLY,
    )

    redoLength = Int32(
        displayedName="Redo length",
        defaultValue=0,
        accessMode=AccessMode.READONLY,
    )


class BaseManualGeometryConfigNode(Configurable):
    offset = make_x_y_offset_node()

    @Slot(
        displayedName="Set manual geometry",
        allowedStates=[State.ACTIVE],
    )
    async def setManual(self):
        await self.get_root()._set_from_manual_config()


class GeometryFileNode(Configurable):
    filePath = String(
        defaultValue="",
        displayedName="File path",
        description="Full path (including filename and suffix) to the desired geometry "
        "file. Keep in mind that the default directory is $KARABO/var/data on device "
        "server node, so it's probably wise to give absolute path.",
        assignment=Assignment.OPTIONAL,
        accessMode=AccessMode.RECONFIGURABLE,
    )
    fileType = String(
        displayedName="File type",
        defaultValue="crystfel",
        description="What kind of file will be loaded. Corresponds to options within "
        "EXtra-geom. Note that the options listed here may not all apply to all "
        "geometries. I think 'crystfel' (uses extra_geom.[geometry type]"
        ".from_crystfel_geom is the most 'universal', so I left that as default.",
        options=["crystfel", "h5", "h5+quadrants"],
        assignment=Assignment.OPTIONAL,
        accessMode=AccessMode.RECONFIGURABLE,
    )
    offset = make_x_y_offset_node()
    updateManualOnLoad = Bool(
        defaultValue=True,
        displayedName="Update manual settings",
        description="If this flag is on, the manual settings on this device will be "
        "updated according to the loaded geometry file. This is useful when you want "
        "to load a file and then tweak the geometry a bit. This will zero current "
        "offset.",
        assignment=Assignment.OPTIONAL,
        accessMode=AccessMode.RECONFIGURABLE,
    )

    @Slot(
        displayedName="Load from file",
        allowedStates=[State.ACTIVE],
    )
    async def loadFromFile(self):
        await self.get_root()._load_from_file()


class ManualGeometryBase(Device):
    __version__ = deviceVersion
    geometry_class = None  # subclass must set
    # subclass must add slot setManual

    availableScenes = VectorString(
        displayedName="Available scenes",
        displayType="Scenes",
        requiredAccessLevel=AccessLevel.OBSERVER,
        accessMode=AccessMode.READONLY,
        defaultValue=[
            "overview",
        ],
        daqPolicy=DaqPolicy.OMIT,
    )

    signalNewGeometry = Signal(String(), String())

    geometryPreview = Image(
        ImageData(np.empty(shape=(0, 0, 0), dtype=np.uint32)),
        displayedName="Geometry preview",
        encoding=EncodingType.RGBA,
        daqPolicy=DaqPolicy.OMIT,
    )

    geometryFile = Node(
        GeometryFileNode,
        displayedName="Geometry file",
        description="Allows loading geometry from CrystFEL geometry file",
    )

    tweakGeometry = Node(
        TweakGeometryNode,
        displayedName="Tweak geometry",
    )

    @Slot(
        displayedName="Send geometry",
        allowedStates=[State.ACTIVE],
        description="Will send 'signalNewGeometry' to connected slots. These will for "
        "example be DetectorAssembler. Note that signal is sent automatically when new "
        "geometry is set - this slot is mostly to be called by manager after "
        "(re)starting assemblers while geometry device is still up."
    )
    async def sendGeometry(self):
        self.signalNewGeometry(
            self.deviceId, geom_utils.serialize_geometry(self.geometry)
        )

    @Slot(
        displayedName="Update preview",
        allowedStates=[State.ACTIVE],
    )
    async def updatePreview(self):
        axis = self.geometry.inspect()
        axis.figure.tight_layout(pad=0)
        axis.figure.set_facecolor("none")
        # axis.figure.set_size_inches(6, 6)
        # axis.figure.set_dpi(300)
        canvas = FigureCanvasAgg(axis.figure)
        canvas.draw()
        image_buffer = np.frombuffer(canvas.buffer_rgba(), dtype=np.uint8).reshape(
            canvas.get_width_height()[::-1] + (4,)
        )
        self.geometryPreview = ImageData(
            image_buffer, encoding=EncodingType.RGBA, bitsPerPixel=3 * 8
        )
        self._set_status("Preview updated")

    async def _load_from_file(self):
        with self.push_state(State.CHANGING):
            geometry = None

            file_path = self.geometryFile.filePath.value
            self._set_status(f"Loading geometry from {file_path}...")
            file_type = self.geometryFile.fileType.value

            try:
                if file_type == "crystfel":
                    geometry = self.geometry_class.from_crystfel_geom(file_path)
                elif file_type == "h5":
                    geometry = self.geometry_class.from_h5_file(file_path)
                elif file_type == "h5+quadrants":
                    geometry = self.geometry_class.from_h5_file_and_quad_positions(
                        file_path,
                        [(Q.x.value, Q.y.value) for Q in self._quadrant_corners],
                    )
                else:
                    raise ValueError(f"Invalid file type {file_type}")
            except FileNotFoundError:
                self._set_status("Geometry file not found", level=logging.WARN)
            except RuntimeError as e:
                self._set_status(
                    f"Failed to load geometry file: {e}", level=logging.WARN
                )
            except Exception as e:
                self._set_status(
                    f"Misc. exception when loading geometry file: {e}",
                    level=logging.WARN,
                )
            else:
                geometry = geometry.offset(
                    (self.geometryFile.offset.x.value, self.geometryFile.offset.y.value)
                )
                await self._set_geometry(geometry)
                self._set_status("Successfully loaded geometry from file")
                if self.geometryFile.updateManualOnLoad.value:
                    self._set_status(
                        "Updating manual settings on device to reflect loaded geometry"
                    )
                    self._update_manual_from_current()
                self.tweakGeometry._reset()
                return True

        return False

    def _update_manual_from_current(self):
        # subclass should implement (neat when loading from CrystFEL geom)
        raise NotImplementedError()

    async def _set_geometry(self, geometry, update_preview=True):
        self.geometry = geometry
        await self.sendGeometry()
        if update_preview:
            await self.updatePreview()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        plt.switch_backend("agg")  # plotting backend which works for preview hack

    async def onInitialization(self):
        self.state = State.INIT
        self.log.INFO("Waiting a second to let slots connect to signal")
        await sleep(1)
        if self.geometryFile.filePath.value and await self._load_from_file():
            ...
        else:
            await self._set_from_manual_config()
        self.state = State.ACTIVE

    @contextlib.contextmanager
    def push_state(self, state):
        previous_state = self.state
        self.state = state
        try:
            yield
        finally:
            self.state = previous_state

    def _set_status(self, text, level=logging.INFO):
        """Add and log a status message.

        Suppresses throttling from the gui server.
        """

        self.status = text
        self.logger.log(level, text)


def make_origin_node(default_x, default_y):
    class OriginNode(BaseManualGeometryConfigNode):
        x = Double(
            defaultValue=default_x,
            accessMode=AccessMode.RECONFIGURABLE,
            assignment=Assignment.OPTIONAL,
        )
        y = Double(
            defaultValue=default_y,
            accessMode=AccessMode.RECONFIGURABLE,
            assignment=Assignment.OPTIONAL,
        )

    return Node(OriginNode)


class ManualOriginGeometryBase(ManualGeometryBase):
    origin = None  # subclass must define

    async def _set_from_manual_config(self):
        self._set_status("Updating geometry from manual configuration")
        geometry = self.geometry_class.from_origin(
            (self.origin.x.value, self.origin.y.value)
        )
        await self._set_geometry(geometry)
        # TODO: allow offset

    @slot
    def requestScene(self, params):
        name = params.get("name", default="overview")
        if name == "overview":
            scene_data = scenes.origin_geometry_overview(
                self.deviceId,
                self.getDeviceSchema(),
            )
            payload = Hash("success", True, "name", name, "data", scene_data)

        return Hash("type", "deviceScene", "origin", self.deviceId, "payload", payload)

    async def _update_manual_from_current(self):
        raise NotImplementedError()


def make_quadrant_corners_node(default_values):
    assert len(default_values) == 4
    assert all(len(x) == 2 for x in default_values)

    class QuadrantCornersNode(BaseManualGeometryConfigNode):
        Q1 = make_x_y_coordinate_node(*default_values[0])
        Q2 = make_x_y_coordinate_node(*default_values[1])
        Q3 = make_x_y_coordinate_node(*default_values[2])
        Q4 = make_x_y_coordinate_node(*default_values[3])

    return Node(QuadrantCornersNode)


class ManualQuadrantsGeometryBase(ManualGeometryBase):
    quadrantCorners = None  # subclass must define (with nice defaults)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._quadrant_corners = [
            self.quadrantCorners.Q1,
            self.quadrantCorners.Q2,
            self.quadrantCorners.Q3,
            self.quadrantCorners.Q4,
        ]

    @slot
    def requestScene(self, params):
        name = params.get("name", default="overview")
        if name == "overview":
            scene_data = scenes.quadrant_geometry_overview(
                self.deviceId,
                self.getDeviceSchema(),
            )
            payload = Hash("success", True, "name", name, "data", scene_data)

        return Hash("type", "deviceScene", "origin", self.deviceId, "payload", payload)

    async def _set_from_manual_config(self):
        self._set_status("Updating geometry from manual configuration")
        with self.push_state(State.CHANGING):
            geometry = self.geometry_class.from_quad_positions(
                [(Q.x.value, Q.y.value) for Q in self._quadrant_corners]
            ).offset(
                (
                    self.quadrantCorners.offset.x.value,
                    self.quadrantCorners.offset.y.value,
                )
            )
            await self._set_geometry(geometry)
            self.tweakGeometry._reset()

    def _update_manual_from_current(self):
        # TODO: consider what to do about offset
        for corner, (x, y) in zip(
            self._quadrant_corners, self.geometry.quad_positions()
        ):
            corner.x = x
            corner.y = y
        self.quadrantCorners.offset.x = 0
        self.quadrantCorners.offset.y = 0


class OrientableModuleListItem(Configurable):
    posX = Double(
        assignment=Assignment.OPTIONAL,
        defaultValue=0,
    )
    posY = Double(
        assignment=Assignment.OPTIONAL,
        defaultValue=0,
    )
    orientX = Int32(assignment=Assignment.OPTIONAL, defaultValue=1)
    orientY = Int32(assignment=Assignment.OPTIONAL, defaultValue=1)


class RotatableModuleListItem(Configurable):
    posX = Double(
        assignment=Assignment.OPTIONAL,
        defaultValue=0,
    )
    posY = Double(
        assignment=Assignment.OPTIONAL,
        defaultValue=0,
    )
    rotate = Int32(assignment=Assignment.OPTIONAL, defaultValue=1)


def make_manual_orientable_module_list_node(defaults):
    class ManualOrientableModuleListNode(BaseManualGeometryConfigNode):
        modules = VectorHash(
            displayedName="Modules",
            rows=OrientableModuleListItem,
            defaultValue=defaults,
            accessMode=AccessMode.RECONFIGURABLE,
            assignment=Assignment.OPTIONAL,
        )

    return Node(ManualOrientableModuleListNode)


class ManualOrientableModuleListGeometryBase(ManualGeometryBase):
    moduleList = None  # subclass must define (with nice defaults)

    @slot
    def requestScene(self, params):
        name = params.get("name", default="overview")
        if name == "overview":
            # Assumes there are correction devices known to manager
            scene_data = scenes.modules_geometry_overview(
                self.deviceId,
                self.getDeviceSchema(),
            )
            payload = Hash("success", True, "name", name, "data", scene_data)

        return Hash("type", "deviceScene", "origin", self.deviceId, "payload", payload)

    async def _set_from_manual_config(self):
        self._set_status("Updating geometry from manual configuration")
        with self.push_state(State.CHANGING):
            geometry = self.geometry_class.from_module_positions(
                [(x, y) for (x, y, _, _) in self.moduleList.modules.value],
                [
                    (orient_x, orient_y)
                    for (_, _, orient_x, orient_y) in self.moduleList.modules.value
                ],
            ).offset((self.moduleList.offset.x.value, self.moduleList.offset.y.value))
            await self._set_geometry(geometry)
            self.tweakGeometry._reset()


def make_manual_rotatable_module_list_node(defaults):
    class ManualRotatableModuleListNode(BaseManualGeometryConfigNode):
        modules = VectorHash(
            displayedName="Modules",
            rows=RotatableModuleListItem,
            defaultValue=defaults,
            accessMode=AccessMode.RECONFIGURABLE,
            assignment=Assignment.OPTIONAL,
        )

    return Node(ManualRotatableModuleListNode)


class ManualRotatableModuleListGeometryBase(ManualGeometryBase):
    moduleList = None  # subclass must define (with nice defaults)

    @slot
    def requestScene(self, params):
        name = params.get("name", default="overview")
        if name == "overview":
            # Assumes there are correction devices known to manager
            scene_data = scenes.modules_geometry_overview(
                self.deviceId,
                self.getDeviceSchema(),
            )
            payload = Hash("success", True, "name", name, "data", scene_data)

        return Hash("type", "deviceScene", "origin", self.deviceId, "payload", payload)

    async def _set_from_manual_config(self):
        self._set_status("Updating geometry from manual configuration")
        with self.push_state(State.CHANGING):
            geometry = self.geometry_class.from_module_positions(
                [(x, y) for (x, y, _) in self.moduleList.modules.value],
                [
                    rotation
                    for (_, _, rotation) in self.moduleList.modules.value
                ],
            ).offset((self.moduleList.offset.x.value, self.moduleList.offset.y.value))
            await self._set_geometry(geometry)
            self.tweakGeometry._reset()
