import enum
import functools
import itertools
import operator
import pathlib

import jinja2
import numpy as np

from . import utils


class BaseKernelRunner:
    _gpu_based = True
    _xp = None  # subclass sets numpy or cupy

    def _pre_init(self):
        # can be used, for example, to import cupy and set as _xp at runtime
        pass

    def _post_init(self):
        # can be used to set GPU / CPU-specific buffers in subclasses
        # without overriding __init__
        pass

    def __init__(
        self,
        pixels_x,
        pixels_y,
        frames,
        constant_memory_cells,
        output_data_dtype=np.float32,
    ):
        self._pre_init()
        self.pixels_x = pixels_x
        self.pixels_y = pixels_y
        self.frames = frames
        if constant_memory_cells == 0:
            # if not set, guess same as input; may save one recompilation
            self.constant_memory_cells = frames
        else:
            self.constant_memory_cells = constant_memory_cells
        self.output_data_dtype = output_data_dtype
        self._post_init()

    @property
    def preview_shape(self):
        return (self.pixels_x, self.pixels_y)

    def correct(self, flags):
        """Correct (already loaded) image data according to flags

        Detector-specific subclass must define this method. It should assume that image
        data, cell table, and other data (including constants) has already been loaded.
        It should probably run some {CPU,GPU} kernel and output should go into
        self.processed_data{,_gpu}.

        Keep in mind that user only gets output from compute_preview or reshape
        (either of these should come after correct).

        The submodules providing subclasses should have some IntFlag enums defining
        which flags are available to pass along to the kernel. A zero flag should allow
        the kernel to do no actual correction - but still copy the data between buffers
        and cast it to desired output type.

        """
        raise NotImplementedError()

    @property
    def preview_data_views(self):
        """Should provide tuple of views of data for different preview layers - as a
        minimum, raw and corrected.  View is expected to have axis order frame, slow
        scan, fast scan with the latter two matching preview_shape (which is X and which
        is Y may differ between detectors)."""

        raise NotImplementedError()

    def flush_buffers(self, constants=None):
        """Optional reset internal buffers (implement in subclasses which need this)"""
        pass

    def compute_previews(self, preview_index):
        """Generate single slice or reduction previews for raw and corrected data and
        any other layers, determined by self.preview_data_views

        Special values of preview_index are -1 for max, -2 for mean, -3 for sum, and
        -4 for stdev (across cells).

        Note that preview_index is taken from data without checking cell table.
        Caller has to figure out which index along memory cell dimension they
        actually want to preview in case it needs to be a specific cell / pulse.

        Will typically reuse data from corrected output buffer. Therefore,
        correct(...) must have been called with the appropriate flags before
        compute_preview(...).
        """
        if preview_index < -4:
            raise ValueError(f"No statistic with code {preview_index} defined")
        elif preview_index >= self.frames:
            raise ValueError(f"Memory cell index {preview_index} out of range")

        if preview_index >= 0:
            fun = operator.itemgetter(preview_index)
        elif preview_index == -1:
            # note: separate from next case because dtype not applicable here
            fun = functools.partial(self._xp.nanmax, axis=0)
        elif preview_index in (-2, -3, -4):
            fun = functools.partial(
                {
                    -2: self._xp.nanmean,
                    -3: self._xp.nansum,
                    -4: self._xp.nanstd,
                }[preview_index],
                axis=0,
                dtype=self._xp.float32,
            )
        # TODO: reuse output buffers
        # TODO: integrate multithreading
        res = (fun(in_buffer_view) for in_buffer_view in self.preview_data_views)
        if self._gpu_based:
            res = (buf.get() for buf in res)
        return res

    def reshape(self, output_order, out=None):
        """Move axes to desired output order"""
        # TODO: avoid copy
        if output_order == self._corrected_axis_order:
            self.reshaped_data = self.processed_data
        else:
            self.reshaped_data = self.processed_data.transpose(
                utils.transpose_order(self._corrected_axis_order, output_order)
            )

        if self._gpu_based:
            return self.reshaped_data.get(out=out)
        else:
            if out is None:
                return self.reshaped_data
            else:
                out[:] = self.reshaped_data


def get_kernel_template(kernel_fn):
    src_dir = pathlib.Path(__file__).absolute().parent
    with (src_dir / "kernels" / kernel_fn).open("r") as fd:
        return jinja2.Template(fd.read())


def grid_to_cover_shape_with_blocks(full_shape, block_shape):
    return tuple(itertools.starmap(utils.ceil_div, zip(full_shape, block_shape)))


class KernelRunnerTypes(enum.Enum):
    GPU = enum.auto()
    CPU = enum.auto()
