# cython: boundscheck=False
# cython: cdivision=True
# cython: wrapararound=False

cdef unsigned char NONE = 0
cdef unsigned char OFFSET = 1
cdef unsigned char REL_GAIN = 2
cdef unsigned char BPMASK = 4
cdef unsigned char STRIXEL = 8

from cython.parallel import prange
from cython.view cimport contiguous
import numpy as np


def correct_burst(
    unsigned short[:, :, :] image_data,
    unsigned char[:, :, :] gain_stage,
    unsigned char[:] cell_table,
    unsigned char flags,
    float[:, :, :, :] offset_map,
    float[:, :, :, :] relgain_map,
    unsigned[:, :, :, :] badpixel_mask,
    float badpixel_fill_value,
    float[:, :, :] output,
):
    cdef int frame, map_cell, x, y
    cdef unsigned char gain
    cdef float corrected
    for frame in prange(image_data.shape[0], nogil=True):
        map_cell = cell_table[frame]
        if map_cell >= offset_map.shape[0]:
            for y in range(image_data.shape[1]):
                for x in range(image_data.shape[2]):
                    output[frame, y, x] = <float>image_data[frame, y, x]
            continue
        for y in range(image_data.shape[1]):
            for x in range(image_data.shape[2]):
                corrected = image_data[frame, y, x]
                gain = gain_stage[frame, y, x]
                # legal values: 0, 1, or 3
                if gain == 2:
                    corrected = badpixel_fill_value
                else:
                    if gain == 3:
                        gain = 2

                    if (flags & BPMASK) and badpixel_mask[map_cell, y, x, gain] != 0:
                        corrected = badpixel_fill_value
                    else:
                        if (flags & OFFSET):
                            corrected = corrected - offset_map[map_cell, y, x, gain]
                        if (flags & REL_GAIN):
                            corrected = corrected / relgain_map[map_cell, y, x, gain]
                output[frame, y, x] = corrected


def correct_single(
    unsigned short[:, :, :] image_data,
    unsigned char[:, :, :] gain_stage,
    unsigned char flags,
    float[:, :, :, :] offset_map,
    float[:, :, :, :] relgain_map,
    unsigned[:, :, :, :] badpixel_mask,
    float badpixel_fill_value,
    float[:, :, :] output,
):
    # ignore "cell table", constant pretends cell 0
    cdef int x, y
    cdef unsigned char gain
    cdef float corrected
    for y in range(image_data.shape[1]):
        for x in range(image_data.shape[2]):
            corrected = image_data[0, y, x]
            gain = gain_stage[0, y, x]
            # legal values: 0, 1, or 3
            if gain == 2:
                corrected = badpixel_fill_value
            else:
                if gain == 3:
                    gain = 2

                if (flags & BPMASK) and badpixel_mask[0, y, x, gain] != 0:
                    corrected = badpixel_fill_value
                else:
                    if (flags & OFFSET):
                        corrected = corrected - offset_map[0, y, x, gain]
                    if (flags & REL_GAIN):
                        corrected = corrected / relgain_map[0, y, x, gain]
            output[0, y, x] = corrected


def strixel_transform(
    float[:, :, ::contiguous] image_data,
    float[:, :, ::contiguous] output
):
    cdef int yin, xin, igap, ichip, xout, yout, frame

    for frame in range(image_data.shape[0]):
        for yin in range(256) :
            yout = int(yin / 3)
            for xin in range(1024) :
                ichip = <int>(xin / 256)
                xout = (ichip * 774) + (xin % 256) * 3 + yin % 3
                # 774 is the chip period, 256*3+6
                output[frame, yout, xout] = image_data[frame, yin, xin]
        # now the gap pixels...
        for yin in range(256):
            yout = <int>(yin / 6) * 2
            for igap in range(3) :
                # first the left side of gap
                xin = igap * 256 + 255
                xout = igap * 774 + 765 + yin % 6
                output[frame, yout, xout] = image_data[frame, yin, xin]
                output[frame, yout+1, xout] = image_data[frame, yin, xin]
                # then the right side is mirrored
                xin = igap * 256 + 255 + 1
                xout = igap * 774 + 765 + 11 - yin % 6
                output[frame, yout, xout] = image_data[frame, yin, xin]
                output[frame, yout+1, xout] = image_data[frame, yin, xin]
