#include <cuda_fp16.h>

{{corr_enum}}

extern "C" {
	__global__ void correct(const unsigned short* data, // shape: memory cell, y, x
	                        const unsigned char* gain_stage, // same shape
	                        const unsigned char* cell_table,
	                        const unsigned char corr_flags,
	                        const float* offset_map,
	                        const float* rel_gain_map,
	                        const unsigned int* bad_pixel_map,
	                        const float bad_pixel_mask_value,
	                        {{output_data_dtype}}* output) {
		const size_t X = {{pixels_x}};
		const size_t Y = {{pixels_y}};
		const size_t input_frames = {{frames}};
		const size_t map_memory_cells = {{constant_memory_cells}};

		const size_t current_frame = blockIdx.x * blockDim.x + threadIdx.x;
		const size_t y = blockIdx.y * blockDim.y + threadIdx.y;
		const size_t x = blockIdx.z * blockDim.z + threadIdx.z;

		if (current_frame >= input_frames || y >= Y || x >= X) {
			return;
		}

		const size_t data_stride_x = 1;
		const size_t data_stride_y = X * data_stride_x;
		const size_t data_stride_frame = Y * data_stride_y;
		const size_t data_index = current_frame * data_stride_frame +
			y * data_stride_y +
			x * data_stride_x;
		float res = (float)data[data_index];

		// gain mapped constant shape: cell, y, x, gain_level (dim size 3)
		// note: in fixed gain mode, constant still provides data for three stages
		const size_t map_stride_gain = 1;
		const size_t map_stride_x = 3 * map_stride_gain;
		const size_t map_stride_y = X * map_stride_x;
		const size_t map_stride_cell = Y * map_stride_y;

		// TODO: warn user about cell_table value of 255 in either mode
		// note: cell table may contain 255 if data didn't arrive
		{% if burst_mode %}
		// burst mode: "cell 255" will get copied
		// TODO: consider masking "cell 255"
		const size_t map_cell = cell_table[current_frame];
		{% else %}
		// single cell: "cell 255" will get "corrected"
		const size_t map_cell = 0;
		{% endif %}

		if (map_cell < map_memory_cells) {
			unsigned char gain = gain_stage[data_index];
			if (gain == 2) {
				// gain should be read as 0, 1, or 3; value of 2 indicates issue
				res = bad_pixel_mask_value;
			} else {
				if (gain == 3) {
					gain = 2;
				}
				const size_t map_index = map_cell * map_stride_cell +
					y * map_stride_y +
					x * map_stride_x +
					gain * map_stride_gain;
				if ((corr_flags & BPMASK) && bad_pixel_map[map_index]) {
					res = bad_pixel_mask_value;
				} else {
					if (corr_flags & OFFSET) {
						res -= offset_map[map_index];
					}
					if (corr_flags & REL_GAIN) {
						res /= rel_gain_map[map_index];
					}
				}
			}
		}

		{% if output_data_dtype == "half" %}
		output[data_index] = __float2half(res);
		{% else %}
		output[data_index] = ({{output_data_dtype}})res;
		{% endif %}
	}

	__global__ void strixel_transform(const {{output_data_dtype}}* data, // shape: memory cell, y, x
	                                  {{output_data_dtype}}* output) {
		const size_t Xin = {{pixels_x}};
		const size_t Yin = {{pixels_y}};
		const size_t Xout = 3090;
		const size_t Yout = 86;
		const size_t input_frames = {{frames}};

		const size_t current_frame = blockIdx.x * blockDim.x + threadIdx.x;
		// following naming from cython version
		const size_t yin = blockIdx.y * blockDim.y + threadIdx.y;
		size_t xin = blockIdx.z * blockDim.z + threadIdx.z;

		// note: hardcoded limits here as only half of y-axis is used
		if (current_frame >= input_frames || yin >= 256 || xin >= 1024) {
			return;
		}

		// avoid race conditions by only writing these once
		const size_t overwritten_columns[18] = {
			765,
			766,
			767,
			774,
			775,
			776,
			1539,
			1540,
			1541,
			1548,
			1549,
			1550,
			2313,
			2314,
			2315,
			2322,
			2323,
			2324
		};

		const size_t data_stride_x = 1;
		const size_t data_stride_y = Xin * data_stride_x;
		const size_t data_stride_frame = Yin * data_stride_y;

		const size_t output_stride_x = 1;
		const size_t output_stride_y = Xout * output_stride_x;
		const size_t output_stride_frame = Yout * output_stride_y;

		const size_t ichip = xin / 256;
		size_t xout = (ichip * 774) + (xin % 256) * 3 + (yin % 3);
		size_t yout = yin / 3;
		bool will_be_overwritten = false;
		size_t out_index, data_index;
		for (int i=0; i<18; ++i) {
			if (xout == overwritten_columns[i]) {
				will_be_overwritten = true;
			}
		}
		if (!will_be_overwritten) {
			out_index = current_frame * output_stride_frame +
				yout * output_stride_y +
				xout * output_stride_x;
			data_index = current_frame * data_stride_frame +
				yin * data_stride_y +
				xin * data_stride_x;
			output[out_index] = data[data_index];
		}
		if (xin < 3) {
			// reuse for the gap pixel case (see cython version)
			const size_t igap = xin;
			yout = (yin / 6) * 2;

			// left side
			xin = igap * 256 + 255;
			xout = igap * 774 + 765 + yin % 6;
			data_index = current_frame * data_stride_frame +
				yin * data_stride_y +
				xin * data_stride_x;
			out_index = current_frame * output_stride_frame +
				yout * output_stride_y +
				xout * output_stride_x;
			output[out_index] = data[data_index];
			out_index = current_frame * output_stride_frame +
				(yout + 1) * output_stride_y +
				xout * output_stride_x;
			output[out_index] = data[data_index];

			// mirror right side
			xin = igap * 256 + 255 + 1;
			xout = igap * 774 + 765 + 11 - yin % 6;
			data_index = current_frame * data_stride_frame +
				yin * data_stride_y +
				xin * data_stride_x;
			out_index = current_frame * output_stride_frame +
				yout * output_stride_y +
				xout * output_stride_x;
			output[out_index] = data[data_index];
			out_index = current_frame * output_stride_frame +
				(yout + 1) * output_stride_y +
				xout * output_stride_x;
			output[out_index] = data[data_index];
		}
	}
}
