import numpy as np
import pytest

from calng.corrections.DsscCorrection import (
    Constants,
    CorrectionFlags,
    DsscCpuRunner,
    DsscGpuRunner,
)

output_dtype = np.float32
corr_dtype = np.float32
pixels_x = 512
pixels_y = 128
memory_cells = 400
offset_map = (
    np.random.random(size=(pixels_x, pixels_y, memory_cells)).astype(corr_dtype) * 20
)
cell_table = np.arange(memory_cells, dtype=np.uint16)
np.random.shuffle(cell_table)
raw_data = np.random.randint(
    low=0, high=2000, size=(memory_cells, pixels_y, pixels_x), dtype=np.uint16
)


# TODO: gather CPU implementations elsewhere
def correct_cpu(data, cell_table, offset_map):
    corr = np.squeeze(data).astype(corr_dtype, copy=True)
    safe_cell_bool = cell_table < offset_map.shape[-1]
    safe_cell_index = cell_table[safe_cell_bool]
    corr[safe_cell_bool] -= offset_map.transpose()[safe_cell_index]
    return corr.astype(output_dtype, copy=False)


corrected_data = correct_cpu(raw_data, cell_table, offset_map)
only_cast_data = np.squeeze(raw_data).astype(output_dtype)


kernel_runners = [
    runner_class(
        pixels_x,
        pixels_y,
        memory_cells,
        constant_memory_cells=memory_cells,
        output_data_dtype=output_dtype,
    )
    for runner_class in (DsscCpuRunner, DsscGpuRunner)
]


def test_only_cast():
    for runner in kernel_runners:
        runner.load_data(raw_data, cell_table)
        runner.correct(CorrectionFlags.NONE)
        assert np.allclose(runner.reshape(runner._corrected_axis_order), only_cast_data)


def test_correct():
    for runner in kernel_runners:
        runner.load_offset_map(offset_map)
        runner.load_data(raw_data, cell_table)
        runner.correct(CorrectionFlags.OFFSET)
        assert np.allclose(runner.reshape(runner._corrected_axis_order), corrected_data)


def test_correct_oob_cells():
    wild_cell_table = cell_table * 2
    reference = correct_cpu(raw_data, wild_cell_table, offset_map)
    for runner in kernel_runners:
        runner.load_offset_map(offset_map)
        # here, half the cell IDs will be out of bounds
        runner.load_data(raw_data, wild_cell_table)
        # should not crash
        runner.correct(CorrectionFlags.OFFSET)
        # should correct as much as possible
        assert np.allclose(runner.reshape(runner._corrected_axis_order), reference)


def test_reshape():
    kernel_runners[0].processed_data = corrected_data
    kernel_runners[1].processed_data.set(corrected_data)
    for runner in kernel_runners:
        assert np.allclose(
            runner.reshape(output_order="xyf"), corrected_data.transpose()
        )


def test_preview_slice():
    kernel_runners[0].processed_data = corrected_data
    kernel_runners[1].processed_data.set(corrected_data)
    for runner in kernel_runners:
        runner.load_data(raw_data, cell_table)
        preview_raw, preview_corrected = runner.compute_previews(42)
        assert np.allclose(
            preview_raw,
            raw_data[42].astype(np.float32),
        )
        assert np.allclose(
            preview_corrected,
            corrected_data[42].astype(np.float32),
        )


def test_preview_max():
    kernel_runners[0].processed_data = corrected_data
    kernel_runners[1].processed_data.set(corrected_data)
    for runner in kernel_runners:
        # note: in case correction failed, still test this separately
        runner.load_data(raw_data, cell_table)
        preview_raw, preview_corrected = runner.compute_previews(-1)
        assert np.allclose(preview_raw, np.max(raw_data, axis=0).astype(np.float32))
        assert np.allclose(
            preview_corrected, np.max(corrected_data, axis=0).astype(np.float32)
        )


def test_preview_mean():
    kernel_runners[0].processed_data = corrected_data
    kernel_runners[1].processed_data.set(corrected_data)
    for runner in kernel_runners:
        runner.load_data(raw_data, cell_table)
        preview_raw, preview_corrected = runner.compute_previews(-2)
        assert np.allclose(preview_raw, np.nanmean(raw_data, axis=0, dtype=np.float32))
        assert np.allclose(
            preview_corrected, np.nanmean(corrected_data, axis=0, dtype=np.float32)
        )


def test_preview_sum():
    kernel_runners[0].processed_data = corrected_data
    kernel_runners[1].processed_data.set(corrected_data)
    for runner in kernel_runners:
        runner.load_data(raw_data, cell_table)
        preview_raw, preview_corrected = runner.compute_previews(-3)
        assert np.allclose(preview_raw, np.nansum(raw_data, axis=0, dtype=np.float32))
        assert np.allclose(
            preview_corrected, np.nansum(corrected_data, axis=0, dtype=np.float32)
        )


def test_preview_std():
    kernel_runners[0].processed_data = corrected_data
    kernel_runners[1].processed_data.set(corrected_data)
    for runner in kernel_runners:
        runner.load_data(raw_data, cell_table)
        preview_raw, preview_corrected = runner.compute_previews(-4)
        assert np.allclose(preview_raw, np.nanstd(raw_data, axis=0, dtype=np.float32))
        assert np.allclose(
            preview_corrected, np.nanstd(corrected_data, axis=0, dtype=np.float32)
        )


def test_preview_valid_index():
    for runner in kernel_runners:
        with pytest.raises(ValueError):
            runner.compute_previews(-5)
        with pytest.raises(ValueError):
            runner.compute_previews(memory_cells)
