import random
import threading
import time
import timeit

import numpy as np
import pytest
from calng import utils


def test_stacking_buffer_shape():
    original_shape = (1, 2, 3)
    assert utils.stacking_buffer_shape(original_shape, 4, 0) == (4, 1, 2, 3)
    assert utils.stacking_buffer_shape(original_shape, 4, 2) == (1, 2, 4, 3)
    assert utils.stacking_buffer_shape(original_shape, 4, 3) == (1, 2, 3, 4)
    assert utils.stacking_buffer_shape(original_shape, 4, -1) == (1, 2, 3, 4)
    assert utils.stacking_buffer_shape(original_shape, 4, -4) == (4, 1, 2, 3)
    with pytest.raises(np.AxisError):
        utils.stacking_buffer_shape(original_shape, 4, 4)
    with pytest.raises(np.AxisError):
        utils.stacking_buffer_shape(original_shape, 4, -5)


def test_set_on_axis():
    A = np.array([[1, 1, 1, 1, 1], [2, 2, 2, 2, 2], [3, 3, 3, 3, 3]])

    manual = A.copy()
    manual[0] = 0
    fun = A.copy()
    utils.set_on_axis(fun, 0, 0, 0)
    assert np.array_equal(manual, fun)

    manual = A.copy()
    manual[1] = np.arange(5)
    fun = A.copy()
    utils.set_on_axis(fun, np.arange(5), 1, 0)
    assert np.array_equal(manual, fun)

    manual = A.copy()
    manual[:, 1] = 0
    fun = A.copy()
    utils.set_on_axis(fun, 0, 1, 1)
    assert np.array_equal(manual, fun)

    with pytest.raises(IndexError):
        utils.set_on_axis(fun, ..., ..., 3)

    # case triggering obvious bug I had made
    A = np.array([[[1, 2, 3], [4, 5, 6]]])
    manual = A.copy()
    manual[:, 1] = 0
    fun = A.copy()
    utils.set_on_axis(fun, 0, 1, 1)
    assert np.array_equal(manual, fun)


def test_get_c_type():
    assert utils.np_dtype_to_c_type(np.float16) == "half"
    assert utils.np_dtype_to_c_type(np.float32) == "float"
    assert utils.np_dtype_to_c_type(np.float64) == "double"

    assert utils.np_dtype_to_c_type(np.uint8) == "unsigned char"
    assert utils.np_dtype_to_c_type(np.uint16) == "unsigned short"
    assert utils.np_dtype_to_c_type(np.uint32) in ("unsigned", "unsigned int")
    assert utils.np_dtype_to_c_type(np.uint64) == "unsigned long"

    assert utils.np_dtype_to_c_type(np.int8) == "char"
    assert utils.np_dtype_to_c_type(np.int16) == "short"
    assert utils.np_dtype_to_c_type(np.int32) == "int"
    assert utils.np_dtype_to_c_type(np.int64) == "long"


class TestThreadsafeCache:
    def test_arg_key_wrap(self):
        calls = []

        @utils.threadsafe_cache
        def fun(a, b, c=1, d=2, *args, **kwargs):
            calls.append((a, b, c, d, args, kwargs))

        # reordering kwargs /does/ matter because dicts are ordered now
        # (note: functools.lru_cache doesn't sort, claims because of speed)
        fun(1, 2, 3, 4, 5, six=6, seven=7)
        fun(1, 2, 3, 4, 5, seven=7, six=6)
        assert len(calls) == 2, "kwargs order matters"
        calls.clear()

        # reordering kw-style positional args does not matter
        fun(1, 2, 1, 2)
        fun(a=1, c=1, b=2, d=2)
        assert len(calls) == 1, "reordering regular args as kws doesn't matter"
        # and omitting default values does not matter
        fun(b=2, a=1)
        fun(1, 2)
        assert len(calls) == 1, "omitting default args doesn't matter"

    def test_threadsafeness(self):
        # wow, synchronization (presumably) makes this take forever *without* decorator
        from_was_called = []

        base_sleep = 1
        random_sleep = 0.1

        @utils.threadsafe_cache
        def was_called(x):
            time.sleep(random.random() * random_sleep + base_sleep)
            from_was_called.append(x)

        threads = []
        num_threads = 1000
        letters = "abcd"
        start_ts = timeit.default_timer()
        for i in range(num_threads):
            for letter in letters:
                thread = threading.Thread(target=was_called, args=(letter,))
                thread.start()
                threads.append(thread)
        submitted_ts = timeit.default_timer()
        print(f"Right after: {len(from_was_called)}")
        for thread in threads:
            thread.join()
        stop_ts = timeit.default_timer()
        total_time = stop_ts - start_ts
        print(f"After join: {len(from_was_called)}")
        print(f"Time to submit: {submitted_ts - start_ts}")
        print(f"Wait for join: {stop_ts - submitted_ts}")
        print(f"Total: {total_time}")

        # check that function was only called with each letter once
        # this is where the decorator from functools will fail
        assert len(from_was_called) == len(
            letters
        ), "Caching prevents recomputation due to threading"

        # check that the function was not locked too broadly
        # (should run faster than sequential lower bound)
        reasonable_time_to_spawn_thread = 0.45 / 1000
        cutoff = (
            len(letters) * base_sleep + reasonable_time_to_spawn_thread * num_threads
        )
        print(f"Cutoff (sequential lower bound): {cutoff}")
        assert (
            total_time < cutoff
        ), "Locking should not be so broad as to make sequential"
        print(
            f"Each thread would have slept [{base_sleep}, {base_sleep + random_sleep})"
        )

        # check that time doesn't go backwards suddenly
        assert total_time >= base_sleep, "These tests should measure time correctly"
